package com.yonge.audio.analysis.detector;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class PercussionRhythmDetector {
	
	private final static Logger LOGGER = LoggerFactory.getLogger(PercussionRhythmDetector.class);
	
	private float[] energyEnvelope;
	
	private float sampleRate;
	
	private int frameSize;
	
	private final float FACTOR = 4.5f;

	public PercussionRhythmDetector(float[] energyEnvelope, float sampleRate, int frameSize) {
		this.energyEnvelope = energyEnvelope;
		this.sampleRate = sampleRate;
		this.frameSize = frameSize;
	}
	
	private float[] calculateEnergyEnvelopeDifference(float[] energyEnvelope) {
        float[] diff = new float[energyEnvelope.length - 1];
        for (int i = 0; i < diff.length; i++) {
            diff[i] = energyEnvelope[i + 1] - energyEnvelope[i];
        }
        return diff;
    }
    
	private float calculateMean(float[] energyEnvelope) {
        float sum = 0;
        for (float value : energyEnvelope) {
            sum += value;
        }
        return sum / energyEnvelope.length;
    }

	private float calculateStandardDeviation(float[] energyEnvelope, float mean) {
        float sum = 0;
        for (float value : energyEnvelope) {
            sum += (value - mean) * (value - mean);
        }
        return (float) Math.sqrt(sum / energyEnvelope.length);
    }
    
	private List<Integer> detectPercussionRhythm(float[] diff, float mean, float stdDev, float factor) {
        List<Integer> beatPositions = new ArrayList<>();
        float dynamicThreshold = mean + factor * stdDev;
        LOGGER.debug("size:{} dynamicThreshold:{} DIFF:{}", diff.length, dynamicThreshold, Arrays.toString(diff));
        
        for (int i = 1; i < diff.length; i++) {
            if (diff[i] > dynamicThreshold && diff[i - 1] <= dynamicThreshold) {
                beatPositions.add(i - 1);
            }
        }
        return beatPositions;
    }

	private Map<Integer, Double> convertPositionsToTime(List<Integer> positions, int frameSize, float sampleRate) {
		Map<Integer, Double> map = new HashMap<Integer, Double>();

		int intervalTime = 150;
		double lastTime = Math.negateExact(intervalTime);
		
        for (int position : positions) {
            double time = (position * frameSize * 1000) / sampleRate;
            if(time - lastTime >= intervalTime) {
            	map.put(position, time);
            	LOGGER.debug("Detected beats at times (millisecond): {}", time);
            	
            	lastTime = time;
            }
        }
        return map;
    }
    
    public Map<Integer, Double> detect(){
    	
        float[] diff = calculateEnergyEnvelopeDifference(energyEnvelope);
        float mean = calculateMean(diff);
        float stdDev = calculateStandardDeviation(diff, mean);
        
        List<Integer> beatPositions = detectPercussionRhythm(diff, mean, stdDev, FACTOR);
        
        return convertPositionsToTime(beatPositions, frameSize, sampleRate);
    }
}
